# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Account for communication
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Account with ID and Public Key
"""

from .crypto import PublicKey, PrivateKey
from .entity import ID, Entity


class Account(Entity):

    def __init__(self, identifier: ID, public_key: PublicKey):
        """
        Create Account with ID and Public Key

        :param identifier: User ID
        :param public_key: User Public Key
        """
        if identifier.address.network.is_communicator():
            super().__init__(identifier)
            # must verify the ID with meta info before creating an account with meta.key
            self.publicKey = public_key
        else:
            raise ValueError('Account ID error')


class User(Account):

    def __init__(self, identifier: ID, private_key: PrivateKey):
        """
        Create User With ID and Private Key

        :param identifier:  User ID
        :param private_key: User Private key
        """
        if identifier.address.network.is_person():
            super().__init__(identifier, private_key.publicKey)
            self.privateKey = private_key
            self.contacts: list = []
        else:
            raise ValueError('User ID error')

    def addContact(self, contact: ID):
        """
        Add contact by ID

        :param contact: ID
        """
        if not contact.address.network.is_person():
            raise AssertionError('Contact must be a person')
        if contact not in self.contacts:
            self.contacts.append(contact)

    def removeContact(self, contact: ID):
        """
        Remove contact by ID

        :param contact: ID
        """
        if contact in self.contacts:
            self.contacts.remove(contact)

    def hasContact(self, contact: ID) -> bool:
        """
        Check whether contains the contact ID

        :param contact: ID
        :return: True/False
        """
        return contact in self.contacts



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

from .address import Address


class ID(str):
    """
        ID for entity (Account/Group)
        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        data format: "name@address[/terminal]"

        fields:
            name     - entity name, the seed of fingerprint to generate address
            address  - a string to identify an entity
            terminal - entity login resource(device), OPTIONAL
    """

    def __new__(cls, identifier: str='',
                name: str='', address: Address=None, terminal: str=None):
        """
        Create ID object with ID string or name + address

        :param identifier: ID string with format 'name@address/terminal'
        :param name:       A string for ID.name
        :param address:    An Address object for ID.address
        :param terminal:   A string for login point
        :return: ID object
        """
        if identifier:
            # return ID object directly
            if isinstance(identifier, ID):
                return identifier
            # get terminal
            pair = identifier.split('/', 1)
            if len(pair) == 2:
                terminal = pair[1]
            else:
                terminal = None
            # get name & address
            pair = pair[0].split('@', 1)
            if len(pair) == 2:
                name = pair[0]
                address = Address(pair[1])
            else:
                name = ''
                address = Address(pair[0])
        elif address:
            # concatenate ID string
            if name:
                identifier = name + '@' + address
            else:
                identifier = address
            if terminal:
                identifier = identifier + '/' + terminal
        else:
            raise AssertionError('Parameters error')
        # verify ID.address, which number must not be ZERO
        if address.number <= 0:
            raise ValueError('Invalid ID (address) string')
        # new ID(str)
        self = super().__new__(cls, identifier)
        self.name = name
        self.address = address
        self.terminal = terminal
        return self

    @property
    def number(self) -> int:
        return self.address.number


class Entity:
    """
        Entity (Account/Group)
        ~~~~~~~~~~~~~~~~~~~~~~

        Entity with ID and name
    """

    def __init__(self, identifier: ID):
        """
        Create Entity with ID

        :param identifier: User/Group ID
        """
        if identifier.number <= 0:
            raise ValueError('Invalid ID')
        super().__init__()
        self.identifier = identifier
        self.name = identifier.name

    def __str__(self):
        clazz = self.__class__.__name__
        identifier = self.identifier
        network = identifier.address.network
        number = identifier.address.number
        name = self.name
        return '<%s: %s(%d|%d) "%s" />' % (clazz, identifier, network, number, name)

    @property
    def number(self) -> int:
        return self.identifier.address.number



# -*- coding: utf-8 -*-
# ==============================================================================
# MIT License
#
# Copyright (c) 2019 Albert Moky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ==============================================================================

"""
    Account/Group Meta data
    ~~~~~~~~~~~~~~~~~~~~~~~

    data format: {
        version: 1,          // algorithm version
        seed: "moKy",        // user/group name
        key: "{public key}", // PK = secp256k1(SK);
        fingerprint: "..."   // CT = sign(seed, SK);
    }

    algorithm:
        fingerprint = sign(seed, SK);

        CT      = fingerprint; // or key.data
        hash    = ripemd160(sha256(CT));
        code    = sha256(sha256(network + hash)).prefix(4);
        address = base58_encode(network + hash + code);
        number  = uint(code);
"""

from .utils import base64_encode, base64_decode
from .crypto import PublicKey, PrivateKey

from .address import NetworkID, Address
from .entity import ID


class Meta(dict):
    """
        Meta for Identifier
        ~~~~~~~~~~~~~~~~~~~

        @enum MKMMetaVersion

        @abstract Defined for algorithm that generating address.

        @discussion Generate & check ID/Address

            MKMMetaVersion_MKM give a seed string first, and sign this seed to get
            fingerprint; after that, use the fingerprint to generate address.
            This will get a firmly relationship between (username, address & key).

            MKMMetaVersion_BTC use the key data to generate address directly.
            This can build a BTC address for the entity ID (no username).

            MKMMetaVersion_ExBTC use the key data to generate address directly, and
            sign the seed to get fingerprint (just for binding username & key).
            This can build a BTC address, and bind a username to the entity ID.
    """
    Version_MKM = 0x01
    Version_BTC = 0x02
    Version_ExBTC = 0x03
    DefaultVersion = Version_MKM

    def __new__(cls, meta: dict=None,
                seed: str='', key: PublicKey=None, fingerprint: bytes=None, version: chr=DefaultVersion):
        """
        Create meta object with meta info

        :param meta:        A dictionary as meta info
        :param seed:        A string as seed (name)
        :param key:         A public key
        :param fingerprint: A data signed by private keys with seed
        :param version:     Algorithm version
        :return: Meta object
        """
        if meta:
            # return Meta object directly
            if isinstance(meta, Meta):
                return meta
            # get fields from dictionary
            version = int(meta['version'])
            if version == cls.Version_MKM or version == cls.Version_ExBTC:
                seed = meta['seed']
                fingerprint = base64_decode(meta['fingerprint'])
            elif version == cls.Version_BTC:
                seed = ''
                fingerprint = None
            else:
                raise ValueError('unsupported version:', version)
            # public key
            key = PublicKey(meta['key'])
        elif version and key:
            # build meta info
            if version == cls.Version_MKM or version == cls.Version_ExBTC:
                meta = {
                    'version': version,
                    'seed': seed,
                    'key': key,
                    'fingerprint': base64_encode(fingerprint),
                }
            elif version == cls.Version_BTC:
                meta = {
                    'version': version,
                    'key': key,
                }
            else:
                raise ValueError('unsupported version:', version)
        else:
            raise AssertionError('Meta parameters error')
        # check meta version
        if version == cls.Version_MKM or version == cls.Version_ExBTC:
            # verify seed and fingerprint
            if not key.verify(seed.encode('utf-8'), fingerprint):
                raise ValueError('Meta data not math')
        # new Meta(dict)
        self = super().__new__(cls, meta)
        self.version = version
        self.seed = seed
        self.key = key
        self.fingerprint = fingerprint
        return self

    @classmethod
    def generate(cls, seed: str, private_key: PrivateKey, version: chr=DefaultVersion):
        """ Generate meta info with seed and private key """
        if version == cls.Version_MKM or version == cls.Version_ExBTC:
            # generate fingerprint with private key
            fingerprint = private_key.sign(seed.encode('utf-8'))
            dictionary = {
                'version': version,
                'seed': seed,
                'key': private_key.publicKey,
                'fingerprint': base64_encode(fingerprint),
            }
            return Meta(dictionary)
        elif version == cls.Version_BTC:
            dictionary = {
                'version': version,
                'key': private_key.publicKey,
            }
            return Meta(dictionary)
        else:
            raise AssertionError('Invalid version')

    def match_identifier(self, identifier: ID) -> bool:
        """ Check ID with meta info """
        return identifier.name == self.seed and self.match_address(identifier.address)

    def match_address(self, address: Address) -> bool:
        """ Check address with meta info """
        return self.generate_address(address.network) == address

    def generate_identifier(self, network: NetworkID) -> ID:
        """ Generate ID with meta info and network ID """
        address = self.generate_address(network=network)
        return ID(name=self.seed, address=address)

    def generate_address(self, network: NetworkID) -> Address:
        """ Generate address with meta info and network ID """
        if self.version == Meta.Version_MKM:
            return Address.generate(fingerprint=self.fingerprint, network=network)
        elif self.version == Meta.Version_BTC or self.version == Meta.Version_ExBTC:
            return Address.generate(fingerprint=self.key.data, network=network)
